const compose = require('../utils/compose')
const { buildCloudFrontHeaders } = require('../utils/transformers')
const { requestFromEvent } = require('../utils/requests')
const { handleOk } = require('../utils/responses')

const DEFAULT_CORS_OPTIONS = Object.freeze({
  cloudFront: false,
  customHeaders: {},
  methods: [],
  origins: []
})

const parseOrigin = origin => {
  // Do not use url package here, as the origin does not necessarily include a
  // protocol, which would cause parsing to return improper results.
  const regexResult = /^(https?:\/\/)?(.+)$/.exec(origin)

  if (regexResult) {
    return {
      host: regexResult[2],
      protocol: regexResult[1] || null
    }
  }

  return {
    host: null,
    protocol: null
  }
}

const allowMethodsHeaders = ({ methods }) => headers => ({
  ...headers,
  'Access-Control-Allow-Methods': methods.join(',')
})

const allowOriginHeaders = (origin, { origins }) => headers => {
  if (!origin) return headers

  const { host } = parseOrigin(origin)
  const allowAll = origins.includes('**')

  if (!origins.includes(host) && !allowAll) return headers

  return {
    ...headers,
    'Access-Control-Allow-Origin': origin
  }
}

const corsHeadersForRequest = (request, options) => {
  const { cloudFront, customHeaders } = options
  const origin = request.header('origin')
  const headers = compose(
    allowMethodsHeaders(options),
    allowOriginHeaders(origin, options)
  )(customHeaders)

  return options.cloudFront ? buildCloudFrontHeaders(headers) : headers
}

const corsMiddleware = options => next => async lambdaParams => {
  const config = { ...DEFAULT_CORS_OPTIONS, ...options }
  const request = requestFromEvent(lambdaParams.event, config)
  const corsHeaders = corsHeadersForRequest(request, config)

  // Return the whole lambda early when OPTIONS requests..
  if (request.isOptions) {
    const response = handleOk('', corsHeaders, 200, config.cloudFront)
    return lambdaParams.callback(null, response)
  }

  const response = await next(lambdaParams)

  return {
    ...response,
    headers: {
      ...corsHeaders,
      ...response.headers
    }
  }
}

module.exports = corsMiddleware
