const { applyMiddleware } = require('../../middleware')

describe('applyMiddleware', () => {
  const handler = jest.fn(() => ({ statusCode: 200 }))

  describe('with no middleware', () => {
    const handlerWithMiddleware = applyMiddleware()(handler)

    it('simply passes on the arguments', () => {
      expect(handlerWithMiddleware('event', 'context', 'callback')).toEqual({ statusCode: 200 })
      expect(handler).toHaveBeenCalledWith('event', 'context', 'callback')
    })
  })

  describe('with one middleware', () => {
    const handler = jest.fn(({ statusCode }) => ({ statusCode: statusCode + 1 }))
    const middleware = next => args => {
      const result = next({ ...args, event: { statusCode: 302 } })
      result.statusCode++
      return result
    }
    const handlerWithMiddleware = applyMiddleware(middleware)(handler)

    it('passes the arguments through the middleware chain', () => {
      expect(handlerWithMiddleware('event', 'context', 'callback')).toEqual({ statusCode: 304 })
      expect(handler).toHaveBeenCalledWith({ statusCode: 302 }, 'context', 'callback')
    })
  })

  describe('with multiple middleware', () => {
    const handler = jest.fn((event, context, callback) => `${event} ${context}C ${callback}`)
    const firstMiddleware = next => args => next({ ...args, context: `${args.context}A` })
    const secondMiddleware = next => args => next({ ...args, context: `${args.context}B`  })
    const handlerWithMiddleware = applyMiddleware(
      firstMiddleware,
      secondMiddleware
    )(handler)

    it('passes the arguments through the middleware chain', () => {
      expect(handlerWithMiddleware('event', 'context', 'callback')).toEqual('event contextABC callback')
      expect(handler).toHaveBeenCalledWith('event', 'contextAB', 'callback')
    })
  })

  describe('with async middleware', () => {
    const handler = jest.fn((event, context, callback) => `${event} ${context}C ${callback}`)
    const firstMiddleware = next => async args => await next({ ...args, context: `${args.context}A` })
    const secondMiddleware = next => async args => await next({ ...args, context: `${args.context}B`  })
    const handlerWithMiddleware = applyMiddleware(
      firstMiddleware,
      secondMiddleware
    )(handler)

    it('passes the arguments through the middleware chain', async () => {
      expect(await handlerWithMiddleware('event', 'context', 'callback')).toEqual('event contextABC callback')
      expect(handler).toHaveBeenCalledWith('event', 'contextAB', 'callback')
    })
  })

  describe('with an async handler', () => {
    const handler = jest.fn(async () => ({ statusCode: 200 }))
    const middleware = next => async args => {
      const result = await next(args)
      result.statusCode = 304
      return result
    }
    const handlerWithMiddleware = applyMiddleware(middleware)(handler)

    it('simply passes on the arguments', async () => {
      expect(await handlerWithMiddleware('event', 'context', 'callback')).toEqual({ statusCode: 304 })
      expect(handler).toHaveBeenCalledWith('event', 'context', 'callback')
    })
  })
})
