const { corsMiddleware } = require('../../middleware')

// TODO: It shouldn't be tested what the middleware responds, but rather which
// method it uses to generate a response to a certain request. The way the
// answer looks (regular Lambda / CloudFront Lambda) should be an implementation
// detail. It is not possible to test this properly at the moment.
describe('corsMiddleware', () => {
  const context = {}

  describe('with an allowed origin', () => {
    const origin = 'localhost:3001'
    const callback = jest.fn()
    const event = {
      headers: { origin },
      httpMethod: 'OPTIONS'
    }
    const nextMock = jest.fn()
    const middleware = corsMiddleware({ origins: [origin, 'random origin'] })(nextMock)

    it('calls the callback function and includes the origin in the headers', () => {
      middleware({ event, context, callback })

      expect(callback).toHaveBeenCalledTimes(1)
      expect(callback.mock.calls[0]).toMatchSnapshot()
    })
  })

  describe('with a not allowed origin', () => {
    const origin = 'localhost:3001'
    const callback = jest.fn()
    const event = {
      headers: { origin },
      httpMethod: 'OPTIONS'
    }
    const nextMock = jest.fn()
    const middleware = corsMiddleware({ origins: ['random origin'] })(nextMock)

    it('calls the callback function and excludes the origin from the headers', () => {
      middleware({ event, context, callback })

      expect(callback).toHaveBeenCalledTimes(1)
      expect(callback.mock.calls[0]).toMatchSnapshot()
    })
  })

  describe('with all origins permitted', () => {
    const nextMock = jest.fn()
    const middleware = corsMiddleware({ origins: ['**'] })(nextMock)
    const callback = jest.fn()

    beforeEach(() => { callback.mockReset() })

    it('calls the callback function and reflects the origin in the headers', () => {
      const event = {
        headers: { origin: 'whew.lad.nice.origin.tld' },
        httpMethod: 'OPTIONS'
      }

      middleware({ event, context, callback })

      expect(callback).toHaveBeenCalledTimes(1)
      expect(callback.mock.calls[0]).toMatchSnapshot()
    })

    describe('with no origin header set', () => {
      const event = {
        headers: {},
        httpMethod: 'OPTIONS'
      }

      it('calls the callback function and returns the original headers', () => {
        middleware({ event, context, callback })

        expect(callback).toHaveBeenCalledTimes(1)
        expect(callback.mock.calls[0]).toMatchSnapshot()
      })
    })
  })

  describe('with no origin', () => {
    const callback = jest.fn()
    const event = {
      headers: {},
      httpMethod: 'OPTIONS'
    }
    const nextMock = jest.fn()
    const middleware = corsMiddleware({ origins: ['random origin'] })(nextMock)

    it('calls the callback function and excludes the origin from the headers', () => {
      middleware({ event, context, callback })

      expect(callback).toHaveBeenCalledTimes(1)
      expect(callback.mock.calls[0]).toMatchSnapshot()
    })
  })

  describe('with an origin that includes the protocol', () => {
    const origin = 'http://localhost:3001'
    const callback = jest.fn()
    const event = {
      headers: { origin },
      httpMethod: 'OPTIONS'
    }
    const nextMock = jest.fn()
    const middleware = corsMiddleware({ origins: ['localhost:3001'] })(nextMock)

    it('calls the callback function and includes the origin including protocol in the headers', () => {
      middleware({ event, context, callback })

      expect(callback).toHaveBeenCalledTimes(1)
      expect(callback.mock.calls[0]).toMatchSnapshot()
    })
  })

  describe('with a set of methods', () => {
    const origin = 'localhost:3001'
    const callback = jest.fn()
    const event = {
      headers: { origin },
      httpMethod: 'OPTIONS'
    }
    const nextMock = jest.fn()
    const middleware = corsMiddleware({
      methods: ['POST'],
      origins: [origin]
    })(nextMock)

    it('calls the callback function and uses the given set of methods', () => {
      middleware({ event, context, callback })

      expect(callback).toHaveBeenCalledTimes(1)
      expect(callback.mock.calls[0]).toMatchSnapshot()
    })
  })

  describe('with a CloudFront request', () => {
    const origin = 'localhost:3001'
    const callback = jest.fn()
    const event = {
      Records: [{
        cf: {
          request: {
            headers: {
              origin: [{
                key: 'Origin',
                value: origin
              }],
            },
            method: 'OPTIONS'
          }
        }
      }]
    }
    const nextMock = jest.fn()
    const middleware = corsMiddleware({
      cloudFront: true,
      origins: [origin, 'random origin']
    })(nextMock)

    it('calls the callback function and includes the origin in the headers', () => {
      middleware({ event, context, callback })

      expect(callback).toHaveBeenCalledTimes(1)
      expect(callback.mock.calls[0]).toMatchSnapshot()
    })
  })

  describe('with custom headers', () => {
    const origin = 'localhost:3001'
    const callback = jest.fn()
    const event = {
      headers: { origin },
      httpMethod: 'OPTIONS'
    }
    const nextMock = jest.fn()
    const middleware = corsMiddleware({
      customHeaders: { 'Some-Header': 'some value' },
      origins: [origin]
    })(nextMock)

    it('calls the callback function and adds the custom headers', () => {
      middleware({ event, context, callback })

      expect(callback).toHaveBeenCalledTimes(1)
      expect(callback.mock.calls[0]).toMatchSnapshot()
    })
  })

  describe('with a non-options request', () => {
    const origin = 'localhost:3001'
    const callback = jest.fn()
    const event = {
      headers: { origin },
      httpMethod: 'GET'
    }
    const nextMock = jest.fn(() => ({ headers: { 'some-header': 'test' } }))
    const middleware = corsMiddleware({ origins: ['localhost:3001'] })(nextMock)

    it('adds CORS headers to the response', async () => {
      const result = await middleware({ event, context, callback })

      expect(callback).not.toHaveBeenCalled()
      expect(result).toMatchSnapshot()
      expect(nextMock).toHaveBeenCalledWith({ event, context, callback })
    })
  })
})
