const { CloudFrontRequest, LambdaRequest, requestFromEvent } = require('../../utils/requests')

describe('AWSRequest', () => {
  describe('#header', () => {
    it('returns the particular header', () => {
      const request = new LambdaRequest({
        headers: {
          'Accept-Encoding': 'gzip, deflate, br'
        }
      })

      expect(request.header('AccEpt-EncoDinG')).toEqual('gzip, deflate, br')
    })
  })

  describe('#headers', () => {
    it('returns the transformed headers', () => {
      const request = new LambdaRequest({
        headers: {
          'Accept-Encoding': 'gzip, deflate, br'
        }
      })

      expect(request.headers).toEqual({ 'accept-encoding': 'gzip, deflate, br' })
    })
  })

  describe('#isOptions', () => {
    describe('with an options request', () => {
      it('returns true', () => {
        const request = new LambdaRequest({ httpMethod: 'options' })
        expect(request.isOptions).toEqual(true)
      })
    })

    describe('with a non-options request', () => {
      it('returns false', () => {
        const request = new LambdaRequest({ httpMethod: 'get' })
        expect(request.isOptions).toEqual(false)
      })
    })
  })

  describe('#method', () => {
    it('returns the request method', () => {
      const request = new LambdaRequest({ httpMethod: 'get' })
      expect(request.method).toEqual('GET')
    })
  })
})

describe('CloudFrontRequest', () => {
  describe('#headers', () => {
    it('returns the transformed headers', () => {
      const request = new CloudFrontRequest({
        Records: [
          {
            cf: {
              request: {
                headers: {
                  'accept-encoding': [
                    {
                      key: 'Accept-Encoding',
                      value: 'gzip, deflate, br'
                    }
                  ]
                }
              }
            }
          }
        ]
      })

      expect(request.headers).toEqual({ 'accept-encoding': 'gzip, deflate, br' })
    })
  })

  describe('#isOptions', () => {
    describe('with an options request', () => {
      it('returns true', () => {
        const request = new CloudFrontRequest({
          Records: [{
            cf: {
              request: { method: 'options' }
            }
          }]
        })
        expect(request.isOptions).toEqual(true)
      })
    })

    describe('with a non-options request', () => {
      it('returns false', () => {
        const request = new CloudFrontRequest({
          Records: [{
            cf: {
              request: { method: 'get' }
            }
          }]
        })
        expect(request.isOptions).toEqual(false)
      })
    })
  })

  describe('#method', () => {
    it('returns the request method', () => {
      const request = new CloudFrontRequest({
        Records: [{
          cf: {
            request: { method: 'get' }
          }
        }]
      })
      expect(request.method).toEqual('GET')
    })
  })
})

describe('requestFromEvent', () => {
  describe('with a Lambda request', () => {
    it('returns a LambdaRequest instance', () => {
      expect(requestFromEvent({}).constructor).toEqual(LambdaRequest)
    })
  })

  describe('with a CloudFront request', () => {
    const request = requestFromEvent({
      Records: [{
        cf: {
          request: {}
        }
      }]
    }, { cloudFront: true })

    it('returns a CloudFrontRequest instance', () => {
      expect(request.constructor).toEqual(CloudFrontRequest)
    })
  })
})
